import re
from typing import List
from dataclasses import dataclass


FUNCTION_REGEX = re.compile(
    r'^(?P<returnType>[^#\s]+)\s+(?P<functionName>[^\s(]+)\s*\((?P<args>[^)]*).*?\n}\n',
    re.MULTILINE | re.DOTALL
)

ARGS_REGEX = re.compile(
    r'(?P<type>[^\s]+)\s+(?P<name>[^,)]+),*\s*'
)

@dataclass
class Argument:
    type: str
    name: str


@dataclass
class Function:
    name: str
    return_type: str
    args: List[Argument]
    text: str


def parse_arguments(args: str):
    return [Argument(m['type'], m['name']) for m in ARGS_REGEX.finditer(args)]


def parse_functions(shader: str):
    return {
        match['functionName']:
        Function(
            match['functionName'],
            match['returnType'],
            parse_arguments(match['args']),
            match[0]
        )
        for match in FUNCTION_REGEX.finditer(shader)
        if match['functionName'] != 'main'
    }



import ctypes

import OpenGL.GL as gl


TYPE_MAP = {
    'float': {
        'uniform': gl.glUniform1f,
        'ctype': ctypes.c_float,
        'buffer_type': gl.GL_R32F,
        'shader_layout': 'r32f',
        'shader_buffer': 'imageBuffer',
        'shader_store':
            lambda name, args: f'imageStore(result, 0, vec4({name}({args}), vec3(0.0)));'
    },
    'vec2': {
        'uniform': gl.glUniform2f,
        'ctype': (ctypes.c_float * 2),
        'buffer_type': gl.GL_RG32F,
        'shader_layout': 'rg32f',
        'shader_buffer': 'imageBuffer',
        'shader_store':
            lambda name, args: f'imageStore(result, 0, vec4({name}({args}), vec2(0.0)));'
    },
    'vec3': {
        'uniform': gl.glUniform3f,
        'ctype': (ctypes.c_float * 4),
        'buffer_type': gl.GL_RGBA32F,
        'shader_layout': 'rgba32f',
        'shader_buffer': 'imageBuffer',
        'shader_store':
            lambda name, args: f'imageStore(result, 0, vec4({name}({args}), 0.0));'
    },
    'vec4': {
        'uniform': gl.glUniform4f,
        'ctype': (ctypes.c_float * 4),
        'buffer_type': gl.GL_RGBA32F,
        'shader_layout': 'rgba32f',
        'shader_buffer': 'imageBuffer',
        'shader_store':
            lambda name, args: f'imageStore(result, 0,{name}({args}));'
    },
    'int': {
        'uniform': gl.glUniform1i,
        'ctype': ctypes.c_int,
        'buffer_type': gl.GL_R32I,
        'shader_layout': 'r32i',
        'shader_buffer': 'iimageBuffer',
        'shader_store':
            lambda name, args: f'imageStore(result, 0, ivec4({name}({args}), ivec3(0)));'
    },
    'ivec2': {
        'uniform': gl.glUniform2i,
        'ctype': (ctypes.c_int * 2),
        'buffer_type': gl.GL_RG32I,
        'shader_layout': 'rg32i',
        'shader_buffer': 'iimageBuffer',
        'shader_store':
            lambda name, args: f'imageStore(result, 0, ivec4({name}({args}), ivec2(0)));'
    },
    'ivec3': {
        'uniform': gl.glUniform3i,
        'ctype': (ctypes.c_int * 4),
        'buffer_type': gl.GL_RGBA32I,
        'shader_layout': 'rgba32i',
        'shader_buffer': 'iimageBuffer',
        'shader_store':
            lambda name, args: f'imageStore(result, 0, ivec4({name}({args}), 0));'
    },
    'ivec4': {
        'uniform': gl.glUniform4i,
        'ctype': (ctypes.c_int * 4),
        'buffer_type': gl.GL_RGBA32I,
        'shader_layout': 'rgba32i',
        'shader_buffer': 'iimageBuffer',
        'shader_store':
            lambda name, args: f'imageStore(result, 0, {name}({args}));'
    },
    'uint': {
        'uniform': gl.glUniform1ui,
        'ctype': ctypes.c_uint,
        'buffer_type': gl.GL_R32UI,
        'shader_layout': 'r32ui',
        'shader_buffer': 'iimageBuffer',
        'shader_store':
            lambda name, args: f'imageStore(result, 0, ivec4({name}({args}), ivec3(0)));'
    },
    'ivec2': {
        'uniform': gl.glUniform2i,
        'ctype': (ctypes.c_int * 2),
        'buffer_type': gl.GL_RG32I,
        'shader_layout': 'rg32i',
        'shader_buffer': 'iimageBuffer',
        'shader_store':
            lambda name, args: f'imageStore(result, 0, ivec4({name}({args}), ivec2(0)));'
    },
    'ivec3': {
        'uniform': gl.glUniform3i,
        'ctype': (ctypes.c_int * 4),
        'buffer_type': gl.GL_RGBA32I,
        'shader_layout': 'rgba32i',
        'shader_buffer': 'iimageBuffer',
        'shader_store':
            lambda name, args: f'imageStore(result, 0, ivec4({name}({args}), 0));'
    },
    'ivec4': {
        'uniform': gl.glUniform4i,
        'ctype': (ctypes.c_int * 4),
        'buffer_type': gl.GL_RGBA32I,
        'shader_layout': 'rgba32i',
        'shader_buffer': 'iimageBuffer',
        'shader_store':
            lambda name, args: f'imageStore(result, 0, {name}({args}));'
    },
    'bool': {
        'uniform': gl.glUniform1i,
        'ctype': ctypes.c_int,
        'buffer_type': gl.GL_R32I,
        'shader_layout': 'r32i',
        'shader_buffer': 'iimageBuffer',
        'shader_store':
            lambda name, args: f'imageStore(result, 0, ivec4({name}({args}), ivec3(0)));'
    },
    'bvec2': {
        'uniform': gl.glUniform2i,
        'ctype': (ctypes.c_int * 2),
        'buffer_type': gl.GL_RG32I,
        'shader_layout': 'rg32i',
        'shader_buffer': 'iimageBuffer',
        'shader_store':
            lambda name, args: f'imageStore(result, 0, ivec4({name}({args}), ivec2(0)));'
    },
    'bvec3': {
        'uniform': gl.glUniform3i,
        'ctype': (ctypes.c_int * 4),
        'buffer_type': gl.GL_RGBA32I,
        'shader_layout': 'rgba32i',
        'shader_buffer': 'iimageBuffer',
        'shader_store':
            lambda name, args: f'imageStore(result, 0, ivec4({name}({args}), 0));'
    },
    'bvec4': {
        'uniform': gl.glUniform4i,
        'ctype': (ctypes.c_int * 4),
        'buffer_type': gl.GL_RGBA32I,
        'shader_layout': 'rgba32i',
        'shader_buffer': 'iimageBuffer',
        'shader_store':
            lambda name, args: f'imageStore(result, 0, ivec4({name}({args})));'
    },
}



import contextlib

from . import shader_parser
from .graphics_context import GraphicsContext
from .shader_function import ShaderFunction


class ShaderContext:
    def __init__(self, shader_text):
        self.function_data = {
            name: data
            for name, data in shader_parser.parse_functions(shader_text).items()
            if data.return_type != 'void'
        }
        self.graphics_context = GraphicsContext()

    @classmethod
    def from_file(cls, filepath):
        with open(filepath) as fin:
            return cls(fin.read())

    def __enter__(self):
        self.graphics_context.__enter__()
        for name, data in self.function_data.items():
            self.__dict__[name] = ShaderFunction(data)
        return self

    def __exit__(self, *args):
        self.graphics_context.__exit__(*args)



import sdl2


class GraphicsContext:
    def __init__(self):
        self.window = None
        self.context = None

    def __enter__(self):
        sdl2.SDL_Init(sdl2.SDL_INIT_VIDEO)
        self.window = sdl2.SDL_CreateWindow(
            b'Shadertest',
            sdl2.SDL_WINDOWPOS_UNDEFINED,
            sdl2.SDL_WINDOWPOS_UNDEFINED,
            1,
            1,
            sdl2.SDL_WINDOW_OPENGL,
        )
        sdl2.video.SDL_GL_SetAttribute(sdl2.video.SDL_GL_CONTEXT_MAJOR_VERSION, 4)
        sdl2.video.SDL_GL_SetAttribute(sdl2.video.SDL_GL_CONTEXT_MINOR_VERSION, 2)
        sdl2.video.SDL_GL_SetAttribute(
            sdl2.video.SDL_GL_CONTEXT_PROFILE_MASK,
            sdl2.video.SDL_GL_CONTEXT_PROFILE_CORE,
        )
        self.context = sdl2.SDL_GL_CreateContext(self.window)

    def __exit__(self, *args):
        sdl2.SDL_GL_DeleteContext(self.context)
        sdl2.SDL_DestroyWindow(self.window)
        sdl2.SDL_Quit()



import ctypes
import collections

import OpenGL.GL as gl
from OpenGL.GL import shaders

from . import builder
from .type_map import TYPE_MAP


VERTEX_SOURCE = '''
#version 120
void main() {
    gl_Position = gl_Vertex;
}
'''


class ShaderFunction:
    def __init__(self, function_data):
        self.function_data = function_data
        self.gl_vao = self.create_gl_vao()
        self.gl_program = self.create_gl_program()
        self.gl_return_buffer = self.create_gl_buffer()
        self.gl_return_texture = self.create_gl_buffer_texture(self.gl_return_buffer)

    def create_gl_vao(self):
        return gl.glGenVertexArrays(1)

    def create_gl_program(self):
        shader_source = builder.build(self.function_data)
        shader = shaders.compileShader(shader_source, gl.GL_FRAGMENT_SHADER)
        vertex_shader = shaders.compileShader(VERTEX_SOURCE, gl.GL_VERTEX_SHADER)
        program = shaders.compileProgram(vertex_shader, shader)
        return program


    def create_gl_buffer(self):
        buffer = gl.glGenBuffers(1)
        gl.glBindBuffer(gl.GL_TEXTURE_BUFFER, buffer)
        host_buffer = TYPE_MAP[self.function_data.return_type]['ctype']()
        gl.glBufferData(
            gl.GL_TEXTURE_BUFFER,
            ctypes.sizeof(host_buffer),
            ctypes.byref(host_buffer),
            gl.GL_STREAM_READ
        )
        gl.glBindBuffer(gl.GL_TEXTURE_BUFFER, 0)
        return buffer

    def create_gl_buffer_texture(self, buffer):
        texture = gl.glGenTextures(1)
        gl.glBindTexture(gl.GL_TEXTURE_BUFFER, texture)
        buffer_type = TYPE_MAP[self.function_data.return_type]['buffer_type']
        gl.glTexBuffer(gl.GL_TEXTURE_BUFFER, buffer_type, buffer)
        gl.glBindTexture(gl.GL_TEXTURE_BUFFER, 0)
        return texture

    def __call__(self, *args):
        takes = len(self.function_data.args)
        given = len(args)
        if given != takes:
            name = self.function_data.name
            raise TypeError(f'{name} takes {takes} positional argument but {given} were given')

        self.bind(args)
        self.draw()
        rvalue = self.read_return_buffer()
        self.unbind()
        return rvalue

    def bind(self, args):
        gl.glBindVertexArray(self.gl_vao)
        gl.glUseProgram(self.gl_program)
        gl.glBindImageTexture(
            0,
            self.gl_return_texture,
            0,
            gl.GL_FALSE,
            0,
            gl.GL_WRITE_ONLY,
            TYPE_MAP[self.function_data.return_type]['buffer_type']
        )
        gl.glUniform1i(0, 0)
        for i, (arg, value) in enumerate(zip(self.function_data.args, args)):
            if not isinstance(value, collections.abc.Iterable):
                value = (value,)
            TYPE_MAP[arg.type]['uniform'](i + 1, *value)

    def draw(self):
        gl.glDrawArrays(gl.GL_POINTS, 0, 1)

    def read_return_buffer(self):
        host_buffer = TYPE_MAP[self.function_data.return_type]['ctype']()
        gl.glBindBuffer(gl.GL_TEXTURE_BUFFER, self.gl_return_buffer)
        gl.glGetBufferSubData(
            gl.GL_TEXTURE_BUFFER,
            0,
            ctypes.sizeof(host_buffer),
            ctypes.byref(host_buffer)
        )
        gl.glBindBuffer(gl.GL_TEXTURE_BUFFER, 0)

        if self.function_data.return_type == 'bool':
            return bool(host_buffer.value)

        if hasattr(host_buffer, 'value'):
            return host_buffer.value

        if self.function_data.return_type.endswith('3'):
            host_buffer = host_buffer[:3]

        return tuple(host_buffer)

    def unbind(self):
        gl.glBindTexture(gl.GL_TEXTURE_BUFFER, 0)
        gl.glUseProgram(0)
        gl.glBindVertexArray(0)



from setuptools import setup

def readme():
    with open('README.md') as readme_file:
        return readme_file.read()

setup(
    name='shadertest',
    version='0.2.1',
    author='Daniel Stokes',
    author_email='stokes.daniel.j@gmail.com',
    url='https://github.com/Kupoman/shadertest',
    description='A Python library for unit testing GLSL functions',
    long_description=readme(),
    long_description_content_type='text/markdown',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.7',
        'Topic :: Software Development :: Testing',
    ],
    keywords='glsl unit test',
    packages=['shadertest'],
    install_requires=[
        'PyOpenGL',
        'PySDL2',
    ],
)



from .type_map import TYPE_MAP


def build_version():
    return '#version 420'


def build_output_uniforms(function):
    layout = TYPE_MAP[function.return_type]['shader_layout']
    buffer_type = TYPE_MAP[function.return_type]['shader_buffer']
    out_uniform = f'layout({layout}) uniform {buffer_type} result;'
    arg_uniforms = [f'uniform {arg.type} {arg.name};' for arg in function.args]

    return '\n'.join((out_uniform, *arg_uniforms))


def build_args(function):
    return ', '.join([arg.name for arg in function.args])


def build_call(function):
    args = build_args(function)
    return TYPE_MAP[function.return_type]['shader_store'](function.name, args)


def build_main(function):
    return '\n'.join([
        'void main () {'
        f'\t{build_call(function)}'
        '}'
    ])


def build(function):
    shader = '\n'.join([
        build_version(),
        build_output_uniforms(function),
        function.text,
        build_main(function),
    ])
    return shader
